
/* most of this code taken from the FAAD2 project (http://www.audiocoding.com/),
   which is licensed under the GPL:
 */
/* This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>

#include <mp4ff.h>
#include <mp4.h>


uint32_t read_callback(void *user_data, void *buffer, uint32_t length)
{
    return fread(buffer, 1, length, (FILE*)user_data);
}

uint32_t seek_callback(void *user_data, uint64_t position)
{
    return fseek((FILE*)user_data, position, SEEK_SET);
}

int get_aac_track(mp4ff_t *infile)
{
    /* find AAC track */
    int i;
    int numTracks = mp4ff_total_tracks(infile);

    for (i = 0; i < numTracks; i++)
    {
        int rc = mp4ff_get_track_type(infile, i);
        if (rc == 1) // Audio track
            return i;
    }

    /* can't decode this */
    return -1;
}

int mp4_set_metadata(MP4FileHandle file, const char *item, const char *val, size_t v_len)
{
    if (!item || (item && !*item) || !val) return 0;

    if (!stricmp(item, "track") || !stricmp(item, "tracknumber"))
    {
        unsigned __int16 trkn, tot;
        int t1 = 0, t2 = 0;
        sscanf(val, "%d/%d", &t1, &t2);
        trkn = t1, tot = t2;
        if (!trkn) return 1;
        if (MP4SetMetadataTrack(file, trkn, tot)) return 1;
    }
    else if (!stricmp(item, "disc") || !stricmp(item, "disknumber"))
    {
        unsigned __int16 disk, tot;
        int t1 = 0, t2 = 0;
        sscanf(val, "%d/%d", &t1, &t2);
        disk = t1, tot = t2;
        if (!disk) return 1;
        if (MP4SetMetadataDisk(file, disk, tot)) return 1;
    }
    else if (!stricmp(item, "compilation"))
    {
        unsigned __int8 cpil = atoi(val);
        if (!cpil) return 1;
        if (MP4SetMetadataCompilation(file, cpil)) return 1;
    }
    else if (!stricmp(item, "tempo"))
    {
        unsigned __int16 tempo = atoi(val);
        if (!tempo) return 1;
        if (MP4SetMetadataTempo(file, tempo)) return 1;
    }
    else if (!stricmp(item, "artist"))
    {
        if (MP4SetMetadataArtist(file, val)) return 1;
    }
    else if (!stricmp(item, "writer"))
    {
        if (MP4SetMetadataWriter(file, val)) return 1;
    }
    else if (!stricmp(item, "title"))
    {
        if (MP4SetMetadataName(file, val)) return 1;
    }
    else if (!stricmp(item, "album"))
    {
        if (MP4SetMetadataAlbum(file, val)) return 1;
    }
    else if (!stricmp(item, "date") || !stricmp(item, "year"))
    {
        if (MP4SetMetadataYear(file, val)) return 1;
    }
    else if (!stricmp(item, "comment"))
    {
        if (MP4SetMetadataComment(file, val)) return 1;
    }
    else if (!stricmp(item, "genre"))
    {
        if (MP4SetMetadataGenre(file, val)) return 1;
    }
    else if (!stricmp(item, "tool"))
    {
//        if (MP4SetMetadataTool(file, val)) return 1;
    }
    else if (!stricmp(item, "cover"))
    {
//        if (MP4SetMetadataCoverArt(file, val, v_len)) return 1;
    }
    else
    {
//        if (MP4SetMetadataFreeForm(file, (char *)item, (u_int8_t *)val, (u_int32_t)v_len)) return 1;
    }

    return 0;
}

void convert_tags(mp4ff_t *infile, MP4FileHandle MP4hFile)
{
    char *tag = NULL, *item = NULL;
    int k, j;

    j = mp4ff_meta_get_num_items(infile);
    for (k = 0; k < j; k++)
    {
        if (mp4ff_meta_get_by_index(infile, k, &item, &tag))
        {
            if (item != NULL && tag != NULL)
            {
                mp4_set_metadata(MP4hFile, item, tag, strlen(tag));
                //////////////////////
                free(item); item = NULL;
                free(tag); tag = NULL;
            }
        }
    }
}

int remux(char *mp4infile, char *mp4outfile, int argc)
{
    char percents[200];
    int percent, old_percent = -1;

    // infile variables
    mp4ff_t *infile;
    long sampleId, numSamples;
    FILE *mp4File;
    unsigned char *buffer;
    int buffer_size;
    int rc, track, timescale;

    // outfile variables
    MP4FileHandle MP4hFile;
    MP4TrackId MP4track;

    /* initialise the callback structure */
    mp4ff_callback_t *mp4cb = malloc(sizeof(mp4ff_callback_t));

    mp4File = fopen(mp4infile, "rb");
    mp4cb->read = read_callback;
    mp4cb->seek = seek_callback;
    mp4cb->user_data = mp4File;

    infile = mp4ff_open_read(mp4cb);
    if (!infile)
    {
        /* unable to open file */
        fprintf(stderr, "Error opening file: %s\n", mp4infile);
        return 1;
    }

    track = get_aac_track(infile);

    buffer = NULL;
    buffer_size = 0;

    mp4ff_get_decoder_config(infile, track, &buffer, &buffer_size);

    timescale = mp4ff_time_scale(infile, track);

    numSamples = mp4ff_num_samples(infile, track);

    MP4hFile = MP4Create(mp4outfile, 0, 0, 0);
    if ( MP4hFile == MP4_INVALID_FILE_HANDLE ) {
        fprintf(stderr, "MP4Create() failed\n" );
        return 0;
    }

    MP4SetTimeScale ( MP4hFile, 90000 );

    MP4track = MP4AddAudioTrack( MP4hFile, timescale, 1024, MP4_MPEG4_AUDIO_TYPE );

    MP4SetAudioProfileLevel ( MP4hFile, 0x0F );

    MP4SetTrackESConfiguration ( MP4hFile, MP4track, (u_int8_t *)buffer, buffer_size );

    if (buffer)
        free(buffer);

    for (sampleId = 0; sampleId < numSamples; sampleId++)
    {
        /* get acces unit from MP4 file */
        buffer = NULL;
        buffer_size = 0;

        rc = mp4ff_read_sample(infile, track, sampleId, &buffer,  &buffer_size);
        if (rc == 0)
        {
            fprintf(stderr, "Reading from MP4 file failed.\n");
            mp4ff_close(infile);
            free(mp4cb);
            fclose(mp4File);
            return 1;
        }

        MP4WriteSample( MP4hFile, MP4track, buffer, buffer_size, MP4_INVALID_DURATION, 0, 1 );

        if (buffer)
            free(buffer);

        percent = min((int)(((double)sampleId*100)/(double)numSamples), 100);
        if (percent > old_percent)
        {
            old_percent = percent;
            sprintf(percents, "%d%% decoding %s to %s.", percent, mp4infile, mp4outfile);
            fprintf(stderr, "%s\r", percents);
        }
    }

    if (argc == 3)
        convert_tags(infile, MP4hFile);

    MP4Close ( MP4hFile );
    mp4ff_close(infile);

    free(mp4cb);
    fclose(mp4File);

    MP4Optimize(mp4outfile, NULL, 0);

    return 0;
}

int main(int argc, char *argv[])
{
    fprintf(stderr, "%s\n", argv[0]);
    fprintf(stderr, "Convert iTunes Music Store .m4p files to standard unencrypted .mp4 files\n");
    fprintf(stderr, "!!! Only works on files for which you have a valid playback key !!!\n\n");

    if (!(argc == 3 || argc == 4))
    {
        fprintf(stderr, "USAGE: %s [-t] infile.m4p outfile.mp4\n", argv[0]);
        fprintf(stderr, "OPTIONS: -t   don't copy tags\n");
        return 1;
    }

    if (remux(argv[argc-2], argv[argc-1], argc))
    {
        fprintf(stderr, "Error while converting mp4 file.\n");
        return 1;
    }

    return 0;
}
